"""Bladerunner networking functions.

Copyright (c) 2014, Activision Publishing, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Activision Publishing, Inc. nor the names of its
  contributors may be used to endorse or promote products derived from this
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import socket


def can_resolve(target):
    """Tries to look up a hostname then bind to that IP address.

    Args:
        target: a hostname or IP address as a string

    Returns:
        True if the target is resolvable to a valid IP address
    """

    try:
        socket.getaddrinfo(target, None)
        return True
    except socket.error:
        return False


def _quadrant_to_ip(quadrant):
    """Convert a single binary quadrant to string base 10 integer."""

    return str(int(quadrant, 2))


def _quadrant_to_binary(quadrant):
    """Convert a single quandrant to a binary string."""

    ip_as_int = int(str(quadrant))
    if 0 <= ip_as_int <= 255:
        return bin(ip_as_int)[2:].rjust(8, str(0))


def _ip_to_binary(ip_addr):
    """Convert a dotted quad IP address to a binary string representation."""

    full_binary = []
    for quadrant in ip_addr.split("."):
        quad_binary = _quadrant_to_binary(quadrant)

        if not quad_binary:
            return None

        full_binary.append(quad_binary)

    return "{0:b}".format(int("".join(full_binary), 2)).rjust(32, str(0))


def _binary_to_ip(binary):
    """Convert an binary IP address to dotted quads."""

    ip_address = []
    for quadrant in range(0, len(binary), 8):
        ip_address.append(_quadrant_to_ip(binary[quadrant:quadrant + 8]))
    return ".".join(ip_address)


def _subnet_to_binary(subnet):
    """Convert a CIDR-ish subnet to it's binary representation.

    Args:
        subnet: string, something like N.N.N.N/NN or N.N.N.N/N.N.N.N

    Returns:
        a tuple of binary representations, (network, netmask)
    """

    try:
        net, mask = subnet.split("/")
    except ValueError:
        return None, None

    if not "." in mask:
        mask = int(mask)
        if 0 <= mask <= 32:
            binary_mask = ("1" * mask).ljust(32, "0")
        else:
            # invalid mask size, not that a /0 is much safer...
            return None, None
    else:
        binary_mask = _ip_to_binary(mask)

    return _ip_to_binary(net), binary_mask


def ips_in_subnet(subnet):
    """Given a CIDR-ish network address, return all member IPs.

    Args:
        subnet: string, something like N.N.N.N/NN or N.N.N.N/N.N.N.N

    Returns:
        list of IPv4 addresses without masks
    """

    binary_net, binary_mask = _subnet_to_binary(subnet)
    if binary_mask is None or binary_net is None:
        return None

    for i in range(32):
        if binary_mask[i] != "1":
            subnet_range = int("1" * (32 - i), 2)
            network_section = binary_net[:i]
            break
    else:
        # catches providing a /32 subnet mask, just return the network as an IP
        return [_binary_to_ip(binary_net)]

    members = []
    for ip_ in range(1, subnet_range):  # skip the network address
        ip_ = bin(ip_)[2:].rjust(32 - len(network_section), str(0))
        members.append(_binary_to_ip("{0}{1}".format(network_section, ip_)))

    return members



"""Helper functions to run Bladerunner from the command line with argparse.

This file is part of Bladerunner.

Copyright (c) 2014, Activision Publishing, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Activision Publishing, Inc. nor the names of its
  contributors may be used to endorse or promote products derived from this
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import os
import sys
import getpass
import argparse

from bladerunner import __version__, __release_date__
from bladerunner.formatting import csv_results, pretty_results


def cmdline_entry():
    """argparse's main entry point, performs logic and preformatting."""

    settings, parser = setup_argparse(sys.argv[1:])

    if settings.getHelp:
        parser.print_usage()
        print_help()
        raise SystemExit

    commands, settings = get_commands(settings)

    get_servers(settings)

    if settings.printFixed and settings.printCSV or not settings.servers:
        parser.print_usage()
        sys.exit(1)

    settings = get_passwords(argparse_unlisted(settings))

    if settings.ssh_key is not None:
        settings.ssh_key = settings.ssh_key[0]
    if settings.username is not None:
        settings.username = settings.username[0]
    if settings.jump_user:
        settings.jump_user = settings.jump_user[0]
    if settings.debug is None:
        settings.debug = True

    if settings.settingsDebug:
        raise SystemExit(str(settings))

    setup_output_file(settings)

    options = convert_to_options(settings)

    if settings.printCSV or settings.csv_char != ",":
        options['style'] = -1

    return commands, settings.servers, options


def cmdline_exit(results, options):
    """A buffer for selecting the correct output function and exiting.

    Args::

        results: the results dictionary from Bladerunner.run
        options: the options dictionary, uses 'style' key only
    """

    if options["style"] < 0 or options["style"] > 3:
        csv_results(results, options)
    else:
        try:
            pretty_results(results, options)
        except UnicodeEncodeError:
            csv_results(results, options)
    raise SystemExit


def convert_to_options(settings):
    """Converts argparse's namespace into a dictionary. Removes temp keys."""

    return {
        "username": settings.username,
        "jump_user": settings.jump_user,
        "jump_host": settings.jump_host,
        "jump_pass": settings.jump_pass,
        "jump_port": settings.jump_port,
        "debug": settings.debug,
        "delay": settings.delay,
        "output_file": settings.output_file,
        "password": settings.password,
        "second_password": settings.second_password,
        "password_safety": settings.password_safety,
        "ssh_key": settings.ssh_key,
        "style": settings.style,
        "csv_char": settings.csv_char,
        "threads": settings.threads,
        "width": settings.printFixed,
        "extra_prompts": settings.extra_prompts or [],
        "progressbar": True,
        "port": settings.port,
        "unix_line_endings": settings.unix_line_endings,
        "windows_line_endings": settings.windows_line_endings,
    }


def print_help():
    """Overrides argparses's --help, prints usage only."""

    sys.stdout.write("""Note: [COMMAND] is only optional with --file/-f
Options:
-a --ascii\t\t\t\tUse ASCII output with normal results (same as --style=1)
-c --command-timeout=<seconds>\t\tShell timeout between commands (default: 20s)
-T --connection-timeout=<seconds>\tSpecify the SSH timeout (default: 20s)
-C --csv\t\t\t\tOutput in CSV format, not grouped by similarity
-E --csv-separator=<char>\t\tSpecify the seperation character with CSV output
   --debug=<int>\t\t\tSend debugging to stdout, optional int of ssh debug level
-e --end\t\t\t\tSignal the end of flags, useful with --debug or -m ordering
-f --file=<file>\t\t\tLoad commands from a file
-x --fixed\t\t\t\tUse a fixed 80 character width for output
-h --help\t\t\t\tThis help screen
-H --host-file=<file>\t\t\tLoad hosts from a file
-j --jumpbox=<host>\t\t\tUse a jumpbox to intermediary the targets
-P --jumpbox-password=<password>\tSeparate jumpbox password (-P to prompt)
-J --jumpbox-port=<port>\t\tUse a non-standard SSH port for the jumpbox
-U --jumpbox-username=<username>\tJumpbox user name (default: {username})
-m --match=<pattern> [pattern] ...\tMatch additional shell prompts
-n --no-password\t\t\tNo password prompt
-N --no-password-check\t\t\tDon't check if the first login succeeded
-o --output-file=<file>\t\t\tAppend the output to a file rather than stdout
-p --password=<password>\t\tSupply the host password on the command line
-D --port\t\t\t\tUse a non non-standard SSH port for the target hosts
-s --second-password=<password>\t\tSupply a second password (-s to prompt)
-S --style=<int>\t\t\tOutput style (0=default, 1=ASCII, 2=double, 3=rounded)
-k --ssh-key=<file>\t\t\tUse a non-default ssh key
-t --threads=<int>\t\t\tMaximum concurrent threads (default: 100)
-d --time-delay=<seconds>\t\tAdd a time delay between hosts (default: 0s)
-X --unix-line-endings\t\t\tForce the use of \\n for newlines
-u --username=<username>\t\tUse a different user name (default: {username})
-v --version\t\t\t\tDisplays version information
-W --windows-line-endings\t\tForce the use of \\r\\n for newlines\n""".format(
        username=getpass.getuser(),
    ))


def get_commands(settings):
    """Opens the command file from the settings namespace.

    Args:
        settings: the argparse namespace from the cmdline

    Returns:
        tuple of (commands list, settings namespace)
    """

    if settings.command_file:
        settings.servers.insert(0, settings.command)
        settings.command = None

        try:
            command_list = []
            with open(settings.command_file[0]) as command_file:
                for command_line in command_file:
                    command_line = command_line.strip()
                    if not command_line:
                        continue
                    else:
                        command_list.append(command_line)
            commands = command_list
        except IOError:
            raise SystemExit("Could not open file: {0}".format(
                settings.command_file[0]))
    else:
        commands = [settings.command]

    return commands, settings


def get_servers(settings):
    """Checks to see if a file has been passed as a list of servers."""

    servers = []

    if settings.host_file:
        try:
            with open(settings.host_file[0], "r") as serverfile:
                for line in serverfile.readlines():
                    for server in line.split(" "):
                        if server.strip():
                            servers.append(server.strip())
            settings.servers = servers
        except IOError:
            raise SystemExit("Could not open file: {0}".format(
                settings.host_file[0]))

    if len(settings.servers) == 0 or not settings.servers[0]:
        raise SystemExit(print_help())


def get_passwords(settings):
    """Prompt and sets all passwords."""

    if settings.usePassword is True and not settings.password:
        settings.password = getpass.getpass("Password: ")

    if settings.setsecond_password and not settings.second_password:
        settings.second_password = getpass.getpass("Second password: ")
    elif not settings.second_password:
        settings.second_password = settings.password

    if settings.setjumpbox_password and settings.jump_host:
        settings.jump_pass = getpass.getpass("Jumpbox password: ")
    elif not settings.jump_pass:
        settings.jump_pass = settings.password

    return settings


def argparse_unlisted(settings):
    """Argparse likes to nest everything in lists. This undoes that."""

    unlistings = [
        "delay",
        "password",
        "second_password",
        "jump_pass",
        "port",
        "jump_port",
        "debug",
        "output_file",
    ]

    for unlisting in unlistings:
        if isinstance(getattr(settings, unlisting), list):
            setattr(settings, unlisting, getattr(settings, unlisting)[0])

    unlist_from_defaults = [
        ("cmd_timeout", 20),
        ("timeout", 20),
        ("threads", 100),
    ]

    for setting, default in unlist_from_defaults:
        if getattr(settings, setting) != default:
            setattr(settings, setting, getattr(settings, setting)[0])

    if settings.printFixed:
        settings.printFixed = 80
    if settings.ascii:
        settings.style = 1
    if settings.csv_char != ',':
        settings.csv_char = settings.csv_char[0][0]

    return settings


def setup_output_file(settings):
    """Make sure we can write to the output file if that's being used."""

    if settings.output_file:
        # "touch" the file, ensures we can write the output before execution
        try:
            with open(settings.output_file, "a"):
                os.utime(settings.output_file, None)
        except IOError as error:
            raise SystemExit("Could not open output file: {0}".format(error))


def setup_argparse(args):
    """Sets up the parser's arguments."""

    parser = argparse.ArgumentParser(
        prog="bladerunner",
        description="A simple way to run quick audits or push changes.",
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--ascii",
        "-a",
        dest="ascii",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--command-timeout",
        "-c",
        dest="cmd_timeout",
        metavar="SECONDS",
        nargs=1,
        type=int,
        default=20,
    )

    parser.add_argument(
        "--connection-timeout",
        "-T",
        dest="timeout",
        metavar="SECONDS",
        nargs=1,
        type=int,
        default=20,
    )

    parser.add_argument(
        "--csv",
        "-C",
        dest="printCSV",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--csv-separator",
        "-E",
        dest="csv_char",
        metavar="CHAR",
        nargs=1,
        type=str,
        default=",",
    )

    parser.add_argument(
        "--debug",
        dest="debug",
        metavar="INT",
        nargs="?",
        type=int,
        default=False,
    )

    parser.add_argument(
        "--end",
        "--this-is-the-end",
        "-e",
        dest="jim_morrison",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--file",
        "-f",
        dest="command_file",
        metavar="FILE",
        nargs=1,
        default=False,
    )

    parser.add_argument(
        "--fixed",
        "-x",
        dest="printFixed",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--help",
        "-h",
        dest="getHelp",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--host-file",
        "-H",
        dest="host_file",
        metavar="FILE",
        nargs=1,
        default=False,
    )

    parser.add_argument(
        "--jumpbox",
        "-j",
        dest="jump_host",
        metavar="HOST",
    )

    parser.add_argument(
        "--jumpbox-password",
        dest="jump_pass",
        metavar="PASSWORD",
        nargs=1,
        default=False,
    )

    parser.add_argument(
        "--jumpbox-username",
        "-U",
        dest="jump_user",
        metavar="USER",
        nargs=1,
        default=False,
    )

    parser.add_argument(
        "--jumpbox-port",
        "-J",
        dest="jump_port",
        metavar="PORT",
        nargs=1,
        type=int,
        default=22,
    )

    parser.add_argument(
        "--match",
        "-m",
        dest="extra_prompts",
        metavar="PATTERN",
        nargs="+",
    )

    parser.add_argument(
        "--no-password",
        "-n",
        dest="usePassword",
        action="store_false",
        default=True,
    )

    parser.add_argument(
        "--output-file",
        "-o",
        dest="output_file",
        metavar="FILE",
        nargs=1,
        default=False,
    )

    parser.add_argument(
        "--password",
        "-p",
        dest="password",
        metavar="PASSWORD",
        nargs=1,
    )

    parser.add_argument(
        "--port",
        "-D",
        dest="port",
        metavar="PORT",
        nargs=1,
        type=int,
        default=22
    )

    parser.add_argument(
        "-P",
        dest="setjumpbox_password",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-s",
        dest="setsecond_password",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--second-password",
        dest="second_password",
        metavar="PASSWORD",
        nargs=1,
    )

    parser.add_argument(
        "--settings",
        dest="settingsDebug",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--ssh-key",
        "-k",
        dest="ssh_key",
        metavar="FILE",
        nargs=1,
    )

    parser.add_argument(
        "--style",
        "-S",
        dest="style",
        metavar="INT",
        type=int,
        default=0,
    )

    parser.add_argument(
        "--time-delay",
        "-d",
        dest="delay",
        metavar="SECONDS",
        nargs=1,
        type=float,
        default=0,
    )

    parser.add_argument(
        "--threads",
        "-t",
        dest="threads",
        metavar="INT",
        nargs=1,
        type=int,
        default=100,
    )

    parser.add_argument(
        "--username",
        "-u",
        dest="username",
        metavar="USER",
        nargs=1,
    )

    parser.add_argument(
        "--unix-line-endings",
        "-X",
        dest="unix_line_endings",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--no-password-check",
        "-N",
        dest="password_safety",
        action="store_false",
        default=True,
    )

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version="Bladerunner {ver} on Python {pyv}. Released: {date}\n".format(
            ver=__version__,
            pyv="{0}.{1}.{2}".format(*sys.version_info[:3]),
            date=__release_date__,
        ),
    )

    parser.add_argument(
        "--windows-line-endings",
        "-W",
        dest="windows_line_endings",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        dest="command",
        metavar="COMMAND",
        nargs="?",
    )

    parser.add_argument(
        dest="servers",
        metavar="HOST",
        nargs=argparse.REMAINDER,
    )

    settings = parser.parse_args(args)
    return (settings, parser)



"""Bladerunner provides a method of pushing changes or running audits on groups
of similar hosts over SSH using pexpect (http://pexpect.sourceforge.net). Can
be extended to use an intermediary host if there are networking restrictions.

Copyright (c) 2014, Activision Publishing, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Activision Publishing, Inc. nor the names of its
  contributors may be used to endorse or promote products derived from this
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import unicode_literals
import os
import sys
import math
import time
import getpass
import pexpect
import threading
from concurrent.futures import ThreadPoolExecutor

from bladerunner.progressbar import ProgressBar
from bladerunner.networking import can_resolve, ips_in_subnet
from bladerunner.formatting import FakeStdOut, format_line, format_output


class Bladerunner(object):
    """Main logic for the serial execution of commands on hosts.

    Initialized by a dictionary with the following optional keys (defaults)::

        username: string username (None/current user)
        password: string plain text password, if required (None)
        ssh_key: string non-default ssh key file location (None)
        delay: integer in seconds to pause between servers (None)
        extra_prompts: list of strings of additional expect prompts ([])
        width: integer terminal width for output, or it uses all (None/guess)
        jump_host: string hostname of intermediary host (None)
        jump_user: alternate username for jump_host (None)
        jump_password: alternate password for jump_host (None)
        jump_port: SSH port for jump_host (22)
        second_password: an additional different password for commands (None)
        password_safety: check if the first login succeeds first (False)
        port: SSH port for the servers (22)
        cmd_timeout: integer in seconds to wait for commands (20)
        timeout: integer in seconds to wait to connect (20)
        threads: integer number of parallel threads to run (100)
        style: integer for outputting. Between 0-3 are pretty, or CSV (0)
        csv_char: string character to use for CSV results (",")
        progressbar: boolean to declare if we want a progress display (False)
        unix_line_endings: force sending LF as line endings for commands
        windows_line_endings: force sending CRLF as line endings for commands
    """

    def __init__(self, options=None):
        """Fills in the options dictionary with any missing keys."""

        if not options:
            options = {}

        defaults = {
            "cmd_timeout": 20,
            "csv_char": ",",
            "debug": False,
            "delay": None,
            "extra_prompts": [],
            "jump_host": None,
            "jump_password": None,
            "jump_user": None,
            "jump_port": 22,
            "output_file": False,
            "password": None,
            "password_safety": False,
            "port": 22,
            "progressbar": False,
            "second_password": None,
            "ssh_key": None,
            "style": 0,
            "threads": 100,
            "timeout": 20,
            "unix_line_endings": False,
            "username": None,
            "width": None,
            "windows_line_endings": False,
        }

        for key, value in defaults.items():
            if not key in options:
                options[key] = value

        options = _set_shells(options)

        self.options = options

        self.errors = [
            "Did not login correctly (err: -1)",
            "Received unexpected password prompt (err: -2)",
            "Could not resolve host (err: -3)",
            "Permission denied (err: -4)",
            "Password denied (err: -5)",
            "Shell prompt guessing failure (err: -6)",
        ]

        if sys.version_info > (3,):
            self.unicode_chr = chr
        else:
            self.unicode_chr = unichr

        self.progress = None
        self.sshc = None
        self.commands = None
        self.commands_on_servers = None

        if not self.options["windows_line_endings"] and \
          not self.options["unix_line_endings"] and hasattr(os, "uname") and \
          "darwin" in os.uname()[0].lower():
            # Apples have special needs... default them to unix line endings
            self.options["unix_line_endings"] = True

        super(Bladerunner, self).__init__()

    def run(self, commands=None, servers=None, commands_on_servers=None):
        """Executes commands on servers.

        Args::

            commands: a list of strings of commands to run
            servers: a list of strings of hostnames
            commands_on_servers: an optional dictionary used when providing
                                 unique lists of commands per server

        Returns:
            a list of dictionaries with two keys: name, and results. results
            is a list of tuples of commands issued and their replies.
        """

        if not isinstance(servers, list):
            servers = [servers]

        if not isinstance(commands, list):
            commands = [commands]

        servers = self._prep_servers(commands, servers, commands_on_servers)

        if self.options["progressbar"]:
            self.progress = ProgressBar(len(servers), self.options)
            self.progress.setup()

        if self.options["jump_host"]:
            jumpuser = self.options["jump_user"] or self.options["username"]
            (self.sshc, error_code) = self.connect(
                self.options["jump_host"],
                jumpuser,
                self.options["jump_pass"],
                self.options["jump_port"],
            )
            if error_code < 0:
                message = int(math.fabs(error_code)) - 1
                raise SystemExit("Jumpbox Error: {0}".format(
                    self.errors[message]))

        if self.options["delay"] or self.options["jump_host"]:
            results = self._run_serial(servers)
        else:
            results = self._run_parallel(servers)

        if self.options["jump_host"]:
            self.close(self.sshc, True)

        if self.options["progressbar"]:
            self.progress.clear()

        return results

    def _run_thread(self, commands, servers, commands_on_servers, callback):
        """Wrapper function to execute self.run with a callback."""

        results = self.run(commands, servers, commands_on_servers)

        if callback:
            callback(results)

    def run_threaded(self, commands=None, servers=None,
                     commands_on_servers=None, callback=None):
        """Non-blocking call which creates and starts a thread for self.run().

        Args::

            commands: a list of strings of commands to run
            servers: a list of strings of hostnames
            commands_on_servers: an optional dictionary used when providing
                                 unique lists of commands per server
            callback: function that will receive results when run is finished

        Returns:
            the started thread object which is running commands on servers
        """

        thread = threading.Thread(
            target=self._run_thread,
            args=(commands, servers, commands_on_servers, callback),
        )
        thread.start()
        return thread

    def _prep_servers(self, commands, servers, commands_on_servers=None):
        """Checks to see if any of the servers passed are CIDR-ish networks.

        Args::

            commands: list of commands to run
            servers: list of servers to run the commands on
            commands_on_servers: dictionary mapping commands to servers

        Returns:
            list of servers to run on, including any expanded networks
        """

        if commands_on_servers is not None:
            for server, command_list in commands_on_servers.items():
                if not isinstance(command_list, list):
                    command_list = [command_list]

                network_members = ips_in_subnet(server)
                if network_members:
                    commands_on_servers.pop(server)
                    for member in network_members:
                        commands_on_servers[member] = command_list
                else:
                    commands_on_servers[server] = command_list

            expanded_servers = commands_on_servers.keys()
            commands = None
        else:
            expanded_servers = []
            for server in servers:
                network_members = ips_in_subnet(server)
                if network_members:
                    expanded_servers.extend(network_members)
                else:
                    expanded_servers.append(server)

        self.commands = commands
        self.commands_on_servers = commands_on_servers

        return expanded_servers

    def _run_parallel(self, servers):
        """Runs commands on servers in parallel when not using a jumpbox."""

        if self.options["password_safety"]:
            return self._run_parallel_safely(servers)
        else:
            return self._run_parallel_no_check(servers)

    def _run_parallel_no_check(self, servers):
        """Runs all servers in parallel without checking if any succeed first.

        Could potentially insta-lock an LDAP account... but we're called from
        lib here so hopefully people know what they're doing. Command line
        entry will default to _run_parallel_safely(). Also called from _safely.

        Args:
            servers: the list of servers to run
        """

        results = []

        max_threads = self.options["threads"]
        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            for result_dict in executor.map(self._run_single, servers):
                results.append(result_dict)

        return results

    def _run_parallel_safely(self, servers):
        """Runs commands in parallel after checking the success of first login.

        Args:
            servers: the list of servers to run
        """

        results = []
        sshr, error_code = self.connect(
            servers[0],
            self.options["username"],
            self.options["password"],
            self.options['port'],
        )
        if error_code < 0:
            message = int(math.fabs(error_code)) - 1
            results.append({
                "name": servers[0],
                "results": [("login", self.errors[message])],
            })
            return results + self._run_serial(
                servers[1:],
            )
        else:
            results.append(self.send_commands(sshr, servers[0]))
            self.close(sshr, not self.options["jump_host"])
            sshr = None
            if self.options["progressbar"]:
                self.progress.update()

            return results + self._run_parallel_no_check(servers[1:])

    def _run_serial(self, servers):
        """Runs commands on servers in serial after jumpbox."""

        results = []
        for server in servers:
            if self.options["delay"] and servers.index(server) > 0:
                time.sleep(self.options["delay"])
            results.append(self._run_single(server))
        return results

    def _run_single(self, server):
        """Runs commands on a single server."""

        (sshr, error_code) = self.connect(
            server,
            self.options["username"],
            self.options["password"],
            self.options['port'],
        )
        if error_code < 0:
            message = int(math.fabs(error_code)) - 1
            results = {
                "name": server,
                "results": [("login", self.errors[message])],
            }
        else:
            results = self.send_commands(sshr, server)
            self.close(sshr, not self.options["jump_host"])
            sshr = None

        if self.options["progressbar"]:
            self.progress.update()

        return results

    def _send_cmd(self, command, server):
        """Internal method to send a single command to the pexpect object.

        Args::

            command: the command to send
            server: the pexpect object to send to

        Returns:
            The formatted output of the command as a string, or -1 on timeout
        """

        try:
            if self.options["unix_line_endings"]:
                server.send("{0}{1}".format(
                    command,
                    self.unicode_chr(0x000A),
                ))
            elif self.options["windows_line_endings"]:
                server.send("{0}{1}{2}".format(
                    command,
                    self.unicode_chr(0x000D),
                    self.unicode_chr(0x000A),
                ))
            else:
                server.sendline(command)

            cmd_response = server.expect(
                self.options["shell_prompts"] +
                self.options["extra_prompts"] +
                self.options["passwd_prompts"],
                self.options["cmd_timeout"],
            )

            if cmd_response >= (
                len(self.options["shell_prompts"]) +
                len(self.options["extra_prompts"])
            ) and len(self.options["second_password"] or "") > 0:
                server.sendline(self.options["second_password"])
                server.expect(
                    self.options["shell_prompts"] +
                    self.options["extra_prompts"],
                    self.options["cmd_timeout"],
                )
        except (pexpect.TIMEOUT, pexpect.EOF):
            return self._try_for_unmatched_prompt(
                server,
                server.before,
                command,
            )

        return format_output(server.before, command)

    def _try_for_unmatched_prompt(self, server, output, command,
                                  _from_login=False, _attempts_left=3):
        """On command timeout, send newlines to guess the missing shell prompt.

        Args:

            server: the sshc object
            output: the sshc.before after issuing command before the timeout
            command: the command issued that caused the initial timeout
            _from_login: if this is called from the login method, return the
                         (connection, code) tuple, or return formatted_output()
            _attempts_left: internal integer to iterate over this function with

        Returns:
            format_output if it can find a new prompt, or -1 on error
        """

        try:
            # prompt is usually in the last 30 chars of the last line of output
            new_prompt = format_line(output.splitlines()[-1][-30:])
        except IndexError:
            # blank last line could cause an IndexError, should send a newline
            pass
        else:
            # escape regex characters
            replacements = ["\\", "/", ")", "(", "[", "]", "{", "}", " ", "$",
                            "?", ">", "<", "^", ".", "*"]
            for char in replacements:
                new_prompt = new_prompt.replace(char, "\{0}".format(char))

            if new_prompt and new_prompt not in self.options["shell_prompts"]:
                self.options["shell_prompts"].append(new_prompt)

        try:
            server.sendline()
            server.expect(
                self.options["shell_prompts"] +
                self.options["extra_prompts"],
                2,
            )
        except (pexpect.TIMEOUT, pexpect.EOF):
            if _attempts_left:
                return self._try_for_unmatched_prompt(
                    server,
                    server.before,
                    command,
                    _from_login=_from_login,
                    _attempts_left=(_attempts_left - 1),
                )
        else:
            self._push_expect_forward(server)
            if _from_login:
                return (server, 1)
            else:
                return format_output(output, command)

        self.send_interrupt(server)

        if _from_login:
            # if we get here, we tried to guess the prompt by sending enter 3
            # times, but still didn't return to that same shell. Something odd
            # is likely happening on the device that needs manual inspection
            return (None, -6)
        else:
            return -1

    def send_commands(self, server, hostname):
        """Executes the commands on a pexpect object.

        Args::

            server: the pexpect host object
            hostname: the string hostname of the server

        Returns:
            a dictionary with two keys::

                name: string of the server's hostname
                results: a list of tuples with each command and its result
        """

        results = {"name": hostname}
        command_results = []

        if self.commands_on_servers:
            commands = self.commands_on_servers[hostname]
        else:
            commands = self.commands

        for command in commands:
            command_result = self._send_cmd(command, server)
            if not command_result or command_result == "\n":
                command_results.append((
                    command,
                    "no output from: {0}".format(command),
                ))
            elif command_result == -1:
                command_results.append((
                    command,
                    "did not return after issuing: {0}".format(command),
                ))
            else:
                command_results.append((command, command_result))

        results["results"] = command_results
        return results

    def _build_ssh_command(self, target, username, port):
        """Builds the ssh connection command.

        Args::

            target: string hostname to connect to
            username: string username to connect as
            port: integer port number to use

        Returns:
            string ssh command with valid option flags
        """

        # default flags
        flags = ["-p", str(port), "-t"]

        if self.options["ssh_key"] and os.path.isfile(self.options["ssh_key"]):
            flags.extend(["-i", self.options["ssh_key"]])

        debug = self.options["debug"]
        if isinstance(debug, int) and debug > 0:
            flags.append("-{0}".format("v" * debug))

        return "ssh {flags} {user}@{host}".format(
            flags = " ".join(flags),
            user=username,
            host=target,
        )

    def connect(self, target, username, password, port):
        """Connects to a server, maybe from another server.

        Args::

            target: the hostname, as a string
            username: the user we are connecting as
            password: list or string plain text password(s) to try
            port: ssh port number, as integer

        Returns:
            a pexpect object that can be passed back here or to send_commands()
        """

        if not can_resolve(target):
            return (None, -3)

        ssh_cmd = self._build_ssh_command(target, username, port)

        if not self.sshc:
            try:
                sshr = pexpect.spawn(ssh_cmd)

                if self.options["debug"]:
                    sshr.logfile_read = FakeStdOut

                login_response = sshr.expect(
                    self.options["passwd_prompts"] +
                    self.options["shell_prompts"] +
                    self.options["extra_prompts"],
                    self.options["timeout"],
                )

                if self.options["jump_host"]:
                    self.sshc = sshr

                return self._multipass(sshr, password, login_response)
            except (pexpect.TIMEOUT, pexpect.EOF):
                return (None, -1)
        else:
            self.sshc.sendline(ssh_cmd)

            try:
                login_response = self.sshc.expect(
                    self.options["passwd_prompts"] +
                    self.options["shell_prompts"] +
                    self.options["extra_prompts"],
                    self.options["timeout"],
                )
            except (pexpect.TIMEOUT, pexpect.EOF):
                self.send_interrupt(self.sshc)
                return (None, -1)

            if self.sshc.before.find("Permission denied") != -1:
                self.send_interrupt(self.sshc)
                return (None, -4)

            return self._multipass(self.sshc, password, login_response)

    def _multipass(self, sshc, passwords, login_response):
        """Buffer to use multiple passwords if using a list of passwords.

        Args::

            sshc: the pexpect object
            passwords: list or string of passwords to try
            login_response: the pexpect return status integer

        Returns:
            a tuple of the pexpect object and error code, tries to be positive
        """

        if not hasattr(passwords, "__iter__"):
            passwords = [passwords]

        error_code = -1
        for password in passwords:
            sshc_returned, error_code = self.login(sshc, password,
                login_response)
            if sshc_returned and error_code > 0:
                return (sshc_returned, error_code)
        else:
            return (None, error_code)

    def login(self, sshc, password, login_response):
        """Internal method for logging in, used by connect.

        Args::

            sshc: the pexpect object
            password: plain text password to send
            login_response: the pexpect return status integer

        Returns:
            a tuple of the connection object and error code
        """

        passlen = len(self.options["passwd_prompts"])

        if login_response == 0:
            # new identity for known_hosts file
            sshc.sendline("yes")
            try:
                login_response = sshc.expect(
                    self.options["passwd_prompts"] +
                    self.options["shell_prompts"] +
                    self.options["extra_prompts"],
                    self.options["timeout"],
                )
            except (pexpect.TIMEOUT, pexpect.EOF):
                self.send_interrupt(sshc)
                return (None, -1)

        if login_response <= passlen and password:
            # password prompt as expected
            sshc.sendline(password)
            try:
                send_response = sshc.expect(
                    self.options["passwd_prompts"] +
                    self.options["shell_prompts"] +
                    self.options["extra_prompts"],
                    self.options["timeout"],
                )
            except (pexpect.TIMEOUT, pexpect.EOF):
                # guess the shell prompt here, we're potentially logged in
                return self._try_for_unmatched_prompt(
                    sshc,
                    sshc.before,
                    "login",
                    _from_login=True,
                )

            if send_response <= len(self.options["passwd_prompts"]):
                # wrong password, or received another password prompt
                self.send_interrupt(sshc)
                return (sshc, -5)
            else:
                # logged in properly with a password
                return (sshc, 1)

        elif login_response <= passlen and not password:
            # password prompt not expected
            self.send_interrupt(sshc)
            return (None, -2)
        else:
            # logged in without using a password. we could check to see
            # if this was intended or not, but really, the point is we've
            # logged into the box, it's time to issue some commands and GTFO
            return (sshc, 1)

    def send_interrupt(self, sshc):
        """Sends ^c and pushes pexpect forward on the object.

        Args:
            sshc: the pexpect object

        Returns:
            None: the sshc maintains its state and should be ready for use
        """

        if not self.options["jump_host"]:
            return
        try:
            sshc.sendline(self.unicode_chr(0x003))
            sshc.expect(
                self.options["shell_prompts"] +
                self.options["extra_prompts"],
                3,
            )
        except (pexpect.TIMEOUT, pexpect.EOF):
            pass
        self._push_expect_forward(sshc)

    def _push_expect_forward(self, sshc):
        """Moves the expect object forwards.

        Args:
            sshc: the pexpect object you'd like to move up
        """

        try:
            sshc.expect(
                self.options["shell_prompts"] +
                self.options["extra_prompts"],
                2,
            )
        except (pexpect.TIMEOUT, pexpect.EOF):
            pass
        try:
            sshc.expect(
                self.options["shell_prompts"] +
                self.options["extra_prompts"],
                2,
            )
        except (pexpect.TIMEOUT, pexpect.EOF):
            pass

    def close(self, sshc, terminate):
        """Closes a connection object.

        Args::

            sshc: the pexpect object to close
            terminate: a boolean value to terminate all connections or not

        Returns:
            None: the sshc will be at the jumpbox, or the connection is closed
        """

        sshc.sendline("exit")

        if terminate:
            sshc.terminate()
        else:
            try:
                sshc.expect(
                    self.options["shell_prompts"] +
                    self.options["extra_prompts"],
                    self.options["cmd_timeout"],
                )
            except (pexpect.TIMEOUT, pexpect.EOF):
                pass


def _set_shells(options):
    """Set password, shell and extra prompts for the username.

    Args:
        options dictionary with username, jump_user and extra_prompts keys.

    Returns:
        options dictionary with shell_prompts and passwd_prompts keys
    """

    shells = [
        "mysql\\>",
        "ftp\\>",
        "telnet\\>",
        "\\[root\\@.*\\]\\#",
        "root\\@.*\\:\\~\\#",
    ]
    password_shells = ["(yes/no)\\\?", "assword:"]

    if not options["username"]:
        options["username"] = getpass.getuser()

    shells.append("\\[{0}@.*\\]\\$".format(options["username"]))
    shells.append("{0}@.*:~\\$".format(options["username"]))
    password_shells.append("{0}@.*assword\\:".format(options["username"]))
    password_shells.append("{0}\\:".format(options["username"]))

    if options["jump_user"]:
        shells.append("\\[{0}@.*\\]\\$".format(options["jump_user"]))
        shells.append("{0}@.*:~\\$".format(options["jump_user"]))
        password_shells.append("{0}@.*assword:".format(options["jump_user"]))
        password_shells.append("{0}:".format(options["jump_user"]))

    options["shell_prompts"] = shells
    options["passwd_prompts"] = password_shells

    if not isinstance(options["extra_prompts"], list):
        options["extra_prompts"] = [options["extra_prompts"]]

    return options



#coding: utf-8
"""Bladerunner output formatting functions.

Copyright (c) 2014, Activision Publishing, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Activision Publishing, Inc. nor the names of its
  contributors may be used to endorse or promote products derived from this
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import print_function

import os
import re

from bladerunner.progressbar import get_term_width


class FakeStdOut(object):
    """An object to pass to pexpect's debug logger to simulate sys.stdout."""

    @staticmethod
    def write(string):
        """Fake write, use print instead."""

        print(string.decode("utf-8").strip())

    @staticmethod
    def flush():
        """Fake flush, print will flush."""

        pass


def no_empties(input_list):
    """Searches through a list and tosses empty elements."""

    output_list = []
    for item in input_list:
        if item:
            output_list.append(item)
    return output_list


def format_output(output, command):
    """Formatting function to strip colours, remove tabs, etc.

    Args::

        output: the pexpect object's before method after issuing the command
        command: the command last issued

    Returns:
        a (hopefully) nicely formatted string of the command's output
    """

    def cmd_in_line(command, line):
        """Checks for long commands wrapping into the output."""

        if len(command) < 60:
            return False

        # how large of command sections we'll look for.
        size = 30
        cmd_split = [command[i:i + size] for i in range(0, len(command), size)]

        for fraction in cmd_split:
            if line.find(fraction) > -1:
                return True

    output = output.splitlines()
    results = []
    # the first line is the command, the last is /probably/ the prompt
    # there can be cases that disobey this though, like exiting without a \n
    for line in output[1:-1]:
        line = format_line(line)
        if line and not cmd_in_line(command, line):
            results.append(line)
    return "\n".join(results)


def format_line(line):
    """Removes whitespace, weird tabs, etc..."""

    for encoding in ["utf-8", "latin-1"]:
        try:
            line = line.decode(encoding)
        except (UnicodeDecodeError, UnicodeEncodeError):
            pass
        else:
            break
    else:
        return line  # can't decode this, not sure what to do. pass it back

    line = line.strip(os.linesep)  # can't strip new lines enough
    line = line.replace("\r", "")  # no extra carriage returns
    line = re.sub("\033\[[0-9;]+m", "", line)  # no colours
    line = re.sub("\x1b\[[0-9;]+G", "", line)  # no crazy tabs
    line = re.sub("\\x1b\[m\\x0f", "", line)
    line = re.sub("^\s+", "", line)  # no trailing whitespace
    return line


def consolidate(results):
    """Makes a list of servers and replies, consolidates dupes.

    Args:
        results: the results dictionary from Bladerunner.run

    Returns:
        a results dictionary, with a names key instead of name, containing a
        lists of hosts with matching outputs
    """

    finalresults = []
    for server in results:
        for tempserver in finalresults:
            if tempserver["results"] == server["results"]:
                tempserver["names"].append(server["name"])
                break
        else:
            server["names"] = [server["name"]]
            del server["name"]
            finalresults.append(server)

    return finalresults


def csv_results(results, options=None):
    """Prints the results consolidated and in a CSV-ish fashion.

    Args::

        results: the results dictionary from Bladerunner.run
        options: dictionary with optional keys:
            csv_char: a character or string to separate with
    """

    if "csv_char" in options:
        csv_char = options["csv_char"]
    else:
        csv_char = ","

    write("server{csv}command{csv}result\r\n".format(csv=csv_char), options)
    for server in results:
        for command, command_result in server["results"]:
            server_name = server.get("name")
            if not server_name:  # catch for consolidated results
                server_name = " ".join(server.get("names"))

            command_result = "\n".join(no_empties(command_result.split("\n")))
            write(
                (
                    "{name_quote}{name}{name_quote}{csv}{cmd_quote}{command}"
                    "{cmd_quote}{csv}{res_quote}{result}{res_quote}\r\n"
                ).format(
                    name_quote='"' * int(" " in server_name),
                    name=server_name,
                    csv=csv_char,
                    cmd_quote='"' * int(" " in command),
                    command=command,
                    res_quote='"' * int(" " in command_result),
                    result=command_result,
                ),
                options,
            )


def prepare_results(results, options=None):
    """Prepare the results and options dictionary for pretty printing.

    Args::

        results: the bladerunner result dictionary
        options: the bladerunner options dictionary

    Returns:
        a tuple of (results, options) after modifying the keys for printing
    """

    if not options:
        options = {}

    left_len = 0
    already_consolidated = False
    for server in results:
        try:
            if len(str(server["name"])) > left_len:
                left_len = len(str(server["name"]))
        except KeyError:
            # catches passing already consolidated results in
            already_consolidated = True
            for server_name in server["names"]:
                if len(server_name) > left_len:
                    left_len = len(server_name)

    if left_len < 6:
        left_len = 6

    # print characters, defined by options["style"]
    options["chars"] = {
        "top_left": ["┌", "*", "╔", "╭"],
        "top": ["─", "-", "═", "─"],
        "top_right": ["┐", "*", "╗", "╮"],
        "top_down": ["┬", "+", "╦", "┬"],
        "side_left": ["├", "*", "╠", "├"],
        "side": ["│", "|", "║", "│"],
        "middle": ["┼", "+", "╬", "┼"],
        "side_right": ["┤", "*", "╣", "┤"],
        "bot_left": ["└", "*", "╚", "╰"],
        "bot": ["─", "-", "═", "─"],
        "bot_right": ["┘", "*", "╝", "╯"],
        "bot_up": ["┴", "+", "╩", "┴"],
    }

    if not "style" in options or not 3 >= options["style"] >= 0:
        options["style"] = 0

    options["left_len"] = left_len

    try:
        width = options["width"] or get_term_width()
    except KeyError:
        width = get_term_width()
    finally:
        options["width"] = width

    if not already_consolidated:
        results = consolidate(results)

    return (results, options)


def pretty_results(results, options=None):
    """Prints the results in a relatively pretty way.

    Args::

        results: the results dictionary from Bladerunner.run
        options: a dictionary with optional keys.
            style: integer style, from 0-3
            jump_host: the string jumpbox hostname
            width: integer fixed width for output
    """

    results, options = prepare_results(results, options)

    pretty_header(options)

    for result in results:
        _pretty_result(result, options, results)

    write(
        "{left_corner}{left}{up}{right}{right_corner}\n".format(
            left_corner=options["chars"]["bot_left"][options["style"]],
            left=options["chars"]["bot"][options["style"]] * (
                options["left_len"] + 2),
            up=options["chars"]["bot_up"][options["style"]],
            right=options["chars"]["bot"][options["style"]] * (
                options["width"] - options["left_len"] - 5),
            right_corner=options["chars"]["bot_right"][options["style"]],
        ),
        options,
    )


def pretty_header(options):
    """Internal function for printing the header of pretty_results.

    Args::

        options: a dictionary with the following keys:
            width: terminal width, already determined in pretty_results
            chars: the character dictionary map, defined in pretty_results
            left_len: the left side length, defined in pretty_results
            jump_host: a string hostname of the jumpbox (if any)
    """

    jumphost = options.get("jump_host")

    if jumphost:
        write(
            "{l_corner}{left}{down}{right}{down}{jumpbox}{r_corner}\n".format(
                l_corner=options["chars"]["top_left"][options["style"]],
                left=options["chars"]["top"][options["style"]] * (
                    options["left_len"]
                    + 2
                ),
                down=options["chars"]["top_down"][options["style"]],
                right=options["chars"]["top"][options["style"]] * (
                    options["width"]
                    - options["left_len"]
                    - 17
                    - len(jumphost)
                ),
                jumpbox=options["chars"]["top"][options["style"]] * (
                    len(jumphost) + 11
                ),
                r_corner=options["chars"]["top_right"][options["style"]],
            ),
            options,
        )

        write(
            (
                "{side} Server{l_gap} {side} Result{r_gap} {side} Jumpbox: "
                "{jumphost} {side}\n"
            ).format(
                side=options["chars"]["side"][options["style"]],
                l_gap=" " * (options["left_len"] - 6),
                r_gap=" " * (
                    options["width"]
                    - options["left_len"]
                    - 25
                    - len(jumphost)
                ),
                jumphost=jumphost,
            ),
            options,
        )
    else:
        write(
            "{l_corner}{left}{down}{right}{r_corner}\n".format(
                l_corner=options["chars"]["top_left"][options["style"]],
                left=options["chars"]["top"][options["style"]] * (
                    options["left_len"]
                    + 2
                ),
                down=options["chars"]["top_down"][options["style"]],
                right=options["chars"]["top"][options["style"]] * (
                    options["width"]
                    - options["left_len"]
                    - 5
                ),
                r_corner=options["chars"]["top_right"][options["style"]],
            ),
            options,
        )

        write(
            "{side} Server{l_gap} {side} Result{r_gap} {side}\n".format(
                side=options["chars"]["side"][options["style"]],
                l_gap=" " * (options["left_len"] - 6),
                r_gap=" " * (options["width"] - options["left_len"] - 13),
            ),
            options,
        )


def _pretty_result(result, options, consolidated_results):
    """Internal function, ran inside of a loop to print super fancy results.

    Args::

        result: the object iterated over in consolidated_results
        options: the options dictionary from pretty_results
        consolidate_results: the output from consolidate
    """

    result_lines = []
    for command, command_result in result["results"]:
        command_split = no_empties(command_result.split("\n"))
        for command_line in command_split:
            result_lines.append(command_line)

    if len(result_lines or "") > len(result["names"]):
        max_length = len(result_lines)
    else:
        max_length = len(result["names"])

    if consolidated_results.index(result) == 0 and options.get("jump_host"):
        # first split has a bottom up character when using a jumpbox
        write(
            "{l_edge}{left}{middle}{right}{up}{jumpbox}{r_edge}\n".format(
                l_edge=options["chars"]["side_left"][options["style"]],
                left=options["chars"]["top"][options["style"]] * (
                    options["left_len"] + 2),
                middle=options["chars"]["middle"][options["style"]],
                right=options["chars"]["top"][options["style"]] * (
                    options["width"]
                    - options["left_len"]
                    - 17
                    - len(options["jump_host"] or "")
                ),
                up=options["chars"]["bot_up"][options["style"]],
                jumpbox=options["chars"]["top"][options["style"]] * (
                    len(options["jump_host"] or "")
                    + 11
                ),
                r_edge=options["chars"]["side_right"][options["style"]],
            ),
            options,
        )
    else:
        # typical horizontal split
        write(
            "{l_side}{left}{middle}{right}{r_side}\n".format(
                l_side=options["chars"]["side_left"][options["style"]],
                left=options["chars"]["top"][options["style"]] * (
                    options["left_len"] + 2),
                middle=options["chars"]["middle"][options["style"]],
                right=options["chars"]["top"][options["style"]] * (
                    options["width"] - options["left_len"] - 5),
                r_side=options["chars"]["side_right"][options["style"]],
            ),
            options,
        )

    for command in range(max_length):
        # print server name or whitespace, mid mark, and leading space
        try:
            write(
                "{side} {server}{gap} {side} ".format(
                    side=options["chars"]["side"][options["style"]],
                    server=result["names"][command],
                    gap=" " * (options["left_len"] - len(
                        str(result["names"][command]))),
                ),
                options,
            )
        except IndexError:
            write(
                "{side} {gap} {side} ".format(
                    side=options["chars"]["side"][options["style"]],
                    gap=" " * options["left_len"],
                ),
                options,
            )

        # print result line, or whitespace, and side mark
        try:
            write(
                "{result}{gap} {side}\n".format(
                    result=result_lines[command],
                    gap=" " * (
                        options["width"]
                        - options["left_len"]
                        - 7
                        - len(result_lines[command])
                    ),
                    side=options["chars"]["side"][options["style"]],
                ),
                options,
            )
        except IndexError:
            write(
                "{gap} {side}\n".format(
                    gap=" " * (options["width"] - options["left_len"] - 7),
                    side=options["chars"]["side"][options["style"]],
                ),
                options,
            )


def write(string, options):
    """Writes a line of output to either the output file or stdout.

    Args::

        string: the string to write out
        options: the options dictionary, uses 'output_file' key only
    """

    if options.get("output_file"):
        with open(options["output_file"], "a") as outputfile:
            outputfile.write(string)
    else:
        try:
            print(string, end="")
        except UnicodeDecodeError as error:
            double_check = _prompt_for_input_on_error(
                "Errored printing the results. Would you like to "
                "write them to a file somewhere instead? ",
                error,
            )

            if double_check.lower().startswith("y"):
                options["output_file"] = _prompt_for_input_on_error(
                    "File name: ",
                    error,
                )
                return write(string, options)
            else:
                raise error


def _prompt_for_input_on_error(user_msg, error):
    """Prompt the user with a message. If they try to quit, raise the error.

    Args::

        user_msg: string message to display to the user
        error: Exception class to raise if the user sends KeyboardInterrupt

    Returns:
        the user's reply to the string message
    """

    try:
        return input(user_msg)
    except KeyboardInterrupt:
        raise error



#coding: utf-8
"""Progress bar and terminal width functions for Bladerunner.

This file is part of Bladerunner.

Copyright (c) 2014, Activision Publishing, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Activision Publishing, Inc. nor the names of its
  contributors may be used to endorse or promote products derived from this
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import division, unicode_literals

import os
import sys


class ProgressBar(object):
    """A simple textual progress bar."""

    def __init__(self, total_updates, options=None):
        """Initializes the object.

        Args:
            total_updates: an integer of how many times update() will be called
            options: a dictionary of additional options. schema:
                width: an integer for fixed terminal width printing
                style: an integer style, between 0-2
                show_counters: a boolean to declare showing the counters or not
                left_padding: a string to pad the left side of the bar with
                right_padding: a string to pad the right side of the bar with
        """

        self.total = total_updates

        if options is None:
            options = {}

        self.total_width = options.get('width') or get_term_width()

        self.counter = 0  # update counter
        self.chars = {
            "left": ["[", "{", ""],
            "right": ["]", "}", ""],
            "space": [" ", " ", "░"],
            25: ["/", ".", "▒"],
            50: ["-", "-", "▓"],
            75: ["\\", "+", "▊"],
            100: ["=", "*", "█"],
        }

        if 0 <= options.get("style", 0) <= len(self.chars["left"]) - 1:
            self.style = options.get("style", 0)
        else:
            self.style = 0

        self.show_counters = options.get("show_counters")

        self.chars["left"][self.style] = "{0}{1}".format(
            options.get("left_padding", ""),
            self.chars["left"][self.style],
        )

        self.chars["right"][self.style] = "{0}{1}".format(
            self.chars["right"][self.style],
            options.get("right_padding", ""),
        )

        if self.show_counters:
            self.width = self.total_width - (
                (len(str(self.total)) * 2)
                + len(self.chars["left"][self.style])
                + len(self.chars["right"][self.style])
                + 2  # the space and the slash
            )
        else:
            self.width = self.total_width - (
                len(self.chars["left"][self.style])
                + len(self.chars["right"][self.style])
            )

        super(ProgressBar, self).__init__()

    def setup(self):
        """Prints an empty progress bar to the screen."""

        if self.show_counters:
            counter_diff = len(str(self.total)) - len(str(self.counter))
            spaces = self.width + counter_diff
            sys.stdout.write("{left}{space}{right} {count}/{total}".format(
                left=self.chars["left"][self.style],
                space=self.chars["space"][self.style] * spaces,
                right=self.chars["right"][self.style],
                count=self.counter,
                total=self.total,
            ))
        else:
            sys.stdout.write("{left}{space}{right}".format(
                left=self.chars["left"][self.style],
                space=self.chars["space"][self.style] * self.width,
                right=self.chars["right"][self.style],
            ))
        sys.stdout.flush()

    def update(self, increment=1):
        """Updates self.counter by increment and reprints the progress bar."""

        self.counter += increment
        counter_diff = len(str(self.total)) - len(str(self.counter))
        percent = (self.counter / self.total) * (self.width + counter_diff)

        sys.stdout.write("\r{left}{spaces}".format(
            left=self.chars["left"][self.style],
            spaces=self.chars[100][self.style] * int(percent),
        ))

        try:
            if not self.total > self.width * 4:
                halfchar = self.chars[rounded(percent, 50)][self.style]
            else:
                halfchar = self.chars[rounded(percent, 25)][self.style]
        except KeyError:
            halfchar = ""

        if self.show_counters:
            sys.stdout.write("{left}{space}{right} {count}/{total}".format(
                left=halfchar,
                space=self.chars["space"][self.style] * (
                    self.width
                    + counter_diff
                    - int(percent)
                    - len(halfchar)
                ),
                right=self.chars["right"][self.style],
                count=self.counter,
                total=self.total
            ))
        else:
            sys.stdout.write("{left}{space}{right}".format(
                left=halfchar,
                space=self.chars["space"][self.style] * (
                    self.width
                    - int(percent)
                    - len(halfchar)
                ),
                right=self.chars["right"][self.style],
            ))
        sys.stdout.flush()

    def clear(self):
        """Clears the progress bar from the screen and resets the cursor."""

        sys.stdout.write("\r{spaces}".format(spaces=" " * self.total_width))
        sys.stdout.flush()
        sys.stdout.write("\r")
        sys.stdout.flush()


def rounded(number, round_to):
    """Internal function for rounding numbers.

    Args:
        number: an integer
        round_to: an integer want the number to be rounded towards

    Returns:
        number, rounded to the nearest round_to
    """

    return int(round(((number - int(number)) * 100) / round_to) * round_to)


def get_term_width():
    """Tries to get the current terminal width, returns 80 if it cannot."""

    env = os.environ

    def ioctl_try(os_fd):
        """Ask the fcntl module for the window size from the os_fd."""

        try:
            import fcntl
            import termios
            import struct
        except ImportError:
            return None

        try:
            return struct.unpack(
                str("hh"),
                fcntl.ioctl(os_fd, termios.TIOCGWINSZ, "1234")
            )
        except Exception:
            return None

    termsize = ioctl_try(0) or ioctl_try(1) or ioctl_try(2)
    if not termsize:
        try:
            os_fd = os.open(os.ctermid(), os.O_RDONLY)
            termsize = ioctl_try(os_fd)
            os.close(os_fd)
        except Exception:
            pass
    if not termsize:
        try:
            termsize = (env["LINES"], env["COLUMNS"])
        except (IndexError, KeyError):
            termsize = (25, 80)
    return int(termsize[1])


def cmd_line_help(name):
    """Overrides argparse's help."""

    raise SystemExit((
        "{name} -- the simple python progress bar used in Bladerunner.\n"
        "Options:\n"
        "\t-c --count=<int>\t\t\tThe number of updates (default 10)\n"
        "\t-d --delay=<seconds>\t\t\tThe seconds between updates (default 1)\n"
        "\t-h --help\t\t\t\tDisplay this help message and quit\n"
        "\t-H --hide-counters\t\t\tDo not show the counters\n"
        "\t-l --left-padding=<str>\t\t\tPad the left side of the bar\n"
        "\t-r --right-padding=<str>\t\tPad the right side of the bar\n"
        "\t-s --style=<int>\t\t\tUse an alternate style (default 0)\n"
        "\t-w --width=<int>\t\t\tThe total width (default max term width)"
    ).format(
        name=name,
    ))


def cmd_line_arguments(args):
    """Sets up argparse for the command line demo."""

    parser = argparse.ArgumentParser(
        prog="progressbar",
        description="progressbar -- a simple python progress bar",
        add_help=False,
    )

    parser.add_argument(
        "--count",
        "-c",
        dest="count",
        metavar="INT",
        type=int,
        nargs=1,
        default=10,
    )

    parser.add_argument(
        "--delay",
        "-d",
        dest="delay",
        metavar="SECONDS",
        type=float,
        nargs=1,
        default=1,
    )

    parser.add_argument(
        "--style",
        "-s",
        dest="style",
        metavar="INT",
        type=int,
        nargs=1,
        default=0,
    )

    parser.add_argument(
        "--width",
        "-w",
        dest="width",
        metavar="INT",
        type=int,
        nargs=1,
        default=get_term_width(),
    )

    parser.add_argument(
        "--left-padding",
        "-l",
        dest="left_padding",
        metavar="STR",
        type=str,
        nargs=1,
        default="",
    )

    parser.add_argument(
        "--right-padding",
        "-r",
        dest="right_padding",
        metavar="STR",
        type=str,
        nargs=1,
        default="",
    )

    parser.add_argument(
        "--hide-counters",
        "-H",
        dest="show_counters",
        action="store_false",
        default=True,
    )

    parser.add_argument(
        "--help",
        "-h",
        dest="helper",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(args)

    if options.helper:
        cmd_line_help(os.path.splitext(__file__)[0])

    unlistings = [
        "count",
        "delay",
        "style",
        "width",
        "left_padding",
        "right_padding",
    ]

    for unlisting in unlistings:
        if isinstance(getattr(options, unlisting), list):
            setattr(options, unlisting, getattr(options, unlisting)[0])

    return options


if __name__ == "__main__":

    import time
    import argparse

    OPTIONS = cmd_line_arguments(sys.argv[1:])

    PROGRESSBAR = ProgressBar(
        OPTIONS.count,
        {
            "style": OPTIONS.style,
            "width": OPTIONS.width,
            "show_counters": OPTIONS.show_counters,
            "left_padding": OPTIONS.left_padding,
            "right_padding": OPTIONS.right_padding,
        },
    )

    try:
        for _ in range(OPTIONS.count):
            time.sleep(OPTIONS.delay)
            PROGRESSBAR.update()
    except KeyboardInterrupt:
        pass

    sys.stdout.write("\n")



"""Bladerunner's setup.py."""


from setuptools import setup
from setuptools.command.test import test as TestCommand


with open("bladerunner/__init__.py", "r") as openinit:
    for line in openinit.readlines():
        if line.startswith("__version__ ="):
            __version__ = line[14:].replace('"', "").replace('"', "").strip()
            break
    else:
        __version__ = "0.0-version-unknown"


class PyTest(TestCommand):
    """Shim in pytest to be able to use it with setup.py test."""

    def finalize_options(self):
        """Stolen from http://pytest.org/latest/goodpractises.html."""

        TestCommand.finalize_options(self)
        self.test_args = ["-v", "-rf", "test"]
        self.test_suite = True

    def run_tests(self):
        """Also shamelessly stolen."""

        # have to import here, outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        raise SystemExit(errno)


setup(
    name="bladerunner",
    version=__version__,
    author="Adam Talsma",
    author_email="adam@demonware.net",
    packages=["bladerunner"],
    install_requires=["pexpect >= 3.0", "futures"],
    scripts=["bin/bladerunner"],
    url="https://github.com/Demonware/bladerunner",
    description="Execution of commands on hosts",
    long_description=(
        "Bladerunner provides an easy to use interface to quickly audit or "
        "push changes to a multitude of hosts. It uses pexpect, so pattern "
        "matching is at its heart. Be aware of returning to a shell prompt "
        "after every command executed. Several options are available for "
        "custom networking and host setups."
    ),
    download_url="https://github.com/Demonware/bladerunner",
    tests_require=["pytest", "mock", "tornado"],
    cmdclass={"test": PyTest},
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: System :: Clustering",
        "Topic :: System :: Systems Administration",
    ],
)
